;===============================================================================
; Implementation of AES (Rijndael) 128 bit on PIC16F84                         ;
;===============================================================================

	LIST	P=PIC16F84A
	INCLUDE P16F84A.INC
	__CONFIG B'11111111110001'

;===============================================================================
; FSR Definition
;===============================================================================
status	equ	0x03
pcl	equ	0x02
pclath	equ	0x0a
eecon1	equ	0x88
fsr	equ	0x04
indf	equ	0x00
porta 	equ	0x05
trisa	equ	0x85
roption equ	0x81
intcon	equ	0x0b
;===============================================================================
; Bit Definition
;===============================================================================
z	equ	0x02
c	equ	0x00
w	equ	0x00
f	equ	0x01
rp0	equ	0x05
rd	equ	0x00
;===============================================================================
; GPR Definition
;===============================================================================
tmp	equ	0x0c
tmp1	equ	0x0d
tmpc	equ	0x0e
lcnt	equ	0x0f
cnt		equ 0x44
delay	equ	0x45
txreg	equ	0x46
rxreg	equ	0x47

s0	equ	0x10
s1	equ	0x11
s2	equ	0x12
s3	equ	0x13
s4	equ	0x14
s5	equ	0x15
s6	equ	0x16
s7	equ	0x17
s8	equ	0x18
s9	equ	0x19
s10	equ	0x1a
s11	equ	0x1b
s12	equ	0x1c
s13	equ	0x1d
s14	equ	0x1e
s15	equ	0x1f

k0	equ	0x20	; Key Organized as follows
k1	equ	0x21	;
k2	equ	0x22	; k0  k1  k2  k3
k3	equ	0x23    ; k4  k5  k6  k7
k4	equ	0x24    ; k8  k9  k10 k11
k5	equ	0x25    ; k12 k13 k14 k15
k6	equ	0x26
k7	equ	0x27
k8	equ	0x28
k9	equ	0x29
k10	equ	0x2a
k11	equ	0x2b
k12	equ	0x2c
k13	equ	0x2d
k14	equ	0x2e
k15	equ	0x2f

t0	equ	0x30
t1	equ	0x31
t2	equ	0x32
t3	equ	0x33
t4	equ	0x34
t5	equ	0x35
t6	equ	0x36
t7	equ	0x37
t8	equ	0x38
t9	equ	0x39
t10	equ	0x3a
t11	equ	0x3b
t12	equ	0x3c
t13	equ	0x3d
t14	equ	0x3e
t15	equ	0x3f

kcnt	equ	0x40
a1	equ	0x41
a2	equ	0x42
a3	equ	0x43
;===========================================================================
;bank and page
;===========================================================================
bank0 MACRO 
	bcf status,rp0
	ENDM
bank1 MACRO
	bsf status,rp0
	ENDM
page0 MACRO
	clrf pclath
	ENDM
page1 MACRO
	bsf pclath,0
	ENDM
;===============================================================================
; Macro to perform registy to registry copy
;===============================================================================
movff	MACRO	a,b
	movf	b,w
	movwf	a
	ENDM
;===========================================================================
;macro for sube byte
;===========================================================================
subbyte MACRO a
	movf a,w
	call lookup
	movwf a
	ENDM
;===============================================================================
; Macro to load a registry by an immediate value
;===============================================================================
movlf	MACRO	c,d
	movlw	d
	movwf	c
	ENDM
;===========================================================================
;constant definition
;===========================================================================
#define _tx porta,0
#define _rx porta,1
#define _trig porta,2
;===========================================================================
; Program Main Entrance
;===========================================================================
	org	0x00
portinit bank1
	movlf trisa, b'11111010'
	bank0
	movlf porta, b'11111011'
	clrf intcon

receivedata movlf lcnt,0x10	;set loop counter
		movlf fsr,0x10	;point to address 0x30
startrx call receive	;call the receive routine
		movwf indf	;move the value of add 0x30 to w
		incf fsr,f	;point to the next address
		decfsz lcnt,f	
		goto startrx
aesroute page1 ;jump to page1 
		call initkey ;initialize the key
		call encrypt
		page0 ;jump back to page0
	
		movlf fsr,0x10
		movlf lcnt,0x10
starttx movff w,indf
		call txmit ;transmin routine
		incf fsr,f
		decfsz lcnt,f
		goto starttx
		goto receivedata
;=====================================================================
; Transmit Subroutine
;	Used Register : tx_reg, delay
;=====================================================================
txmit	movwf	txreg		; save in txreg
		bcf	_tx		; send start bit
		movlf	cnt,0x08	; configuration: 9600-8-N-1
		call	rxwait		; delay 104 us
txbit	btfsc	txreg,0		; data transmitted from LSB to MSB
		bsf	_tx
		btfss	txreg,0x00
		bcf	_tx
		rrf	txreg,f		; rotate txreg to prepare next bit transmission
		nop
		call	txwait		; delay 104 us
		decfsz	cnt,f		; repeat this 8 times
		goto	txbit
		bsf	_tx		; send stop bit
		call	txwait		; delay 104 us
		call	txwait		; another delay for synchronization
		return

txwait	movlf	delay,.29	; Routine to generate 104 cycles delay
		goto	txwait1		; 
rxwait	movlf	delay,.32
txwait1	decfsz	delay,f		; loop till delay register clear
		goto	txwait1
		return
;=====================================================================
; Receive Subroutine
;=====================================================================
receive	btfsc	_rx		; wait untuil start bit detected
		goto	receive
		call	rxwait		; wait 104 us
		movlf	cnt,0x08	; configuration : 9600-8-N-1
		clrf	rxreg		; initialize rxreg
rcv1	call	txwait		; wait 104 us
		bcf	rxreg,0x07
		btfsc	_rx		; copy _rx into rxreg bit 7
		bsf	rxreg,0x07
		decfsz	cnt,w
		rrf	rxreg,f		; rotate rxreg, prepare for the next bit
		decfsz	cnt,f
		goto	rcv1
		call	txwait		; delay 104 us
		btfss	_rx		; check stop bit
		goto	rcv2		; invalid stop bit
		movf	rxreg,w
		return

rcv2	movlf	rxreg,0xff	; error in reception
		return
;===============================================================================
; Encryption Routine
;===============================================================================
	org ($+0xff)&~0xff
encrypt	movlf	kcnt,0x8d
	movlf	lcnt,0x08   ;loop for the encryption
	bsf		_trig 		; set the triger. 
	call	addrndk		;
	call 	fSubByte	;
	bcf		_trig  		; stop the trigger.
	call 	fShiftRow	;
	call	mixcol		;
	call 	getk		;
	call	addrndk		; first round.
encra call fSubByte
	call 	fShiftRow
	call	mixcol
	call	getk
	call 	addrndk
	decfsz	lcnt,f
	goto	encra
	call	fSubByte
	call 	fShiftRow
	call	getk
	call	addrndk
	return	; encryption ends here
;==============================================================================
;Routine to initialize the key value.
;==============================================================================
initkey movlf k0,0x0f
		movlf k1,0x32
		movlf k2,0xb0
		movlf k3,0xf2
		movlf k4,0x01
		movlf k5,0x35
		movlf k6,0x67
		movlf k7,0x8b
		movlf k8,0xdd
		movlf k9,0x00
		movlf k10,0x49
		movlf k11,0x1a
		movlf k12,0x7e
		movlf k13,0xff
		movlf k14,0x29
		movlf k15,0x0b
		return
;======================================================================
; Function to XOR s0:s15 with k0:k15
;======================================================================
addrndk	movlf	fsr,k0
adda	movf	indf,w
	bcf	fsr,5
	bsf	fsr,4
	xorwf	indf,f
	bsf	fsr,5
	bcf	fsr,4
	incf	fsr,f
	btfss	fsr,4
	goto	adda
	return
;========================================================================
; Function to generate next subkey
;========================================================================
getk	movlf	pclath,sbox
	movf	k13,w
	call	lookup
	xorwf	k0,f
	movf	k14,w
	call	lookup
	xorwf	k1,f
	movf	k15,w
	call	lookup
	xorwf	k2,f
	movf	k12,w
	call	lookup
	xorwf	k3,f
	movf	kcnt,w
	call	get2
	movwf	kcnt
	xorwf	k0,f
	movlf	fsr,k0
	movlf	tmp1,0x0c
xora	movff	tmp,indf
	movlw	0x04
	addwf	fsr,f
	movf	tmp,w
	xorwf	indf,f
	movlw	0xfd
	addwf	fsr,f
	decfsz	tmp1,f
	goto	xora
	return
;===========================================================================
;Subbyte routine
;===========================================================================
fSubByte movlf pclath,sbox
	call gsubbyte
	return
;===========================================================================
;routine for inverse subbyte
;===========================================================================
;iSubByte movlf pclath, isbox
;	call gsubbyte
;	return
;===========================================================================
;routine for general subebyte
;===========================================================================
gsubbyte subbyte s0
	subbyte s1
	subbyte s2
	subbyte s3
	subbyte s4
	subbyte s5
	subbyte s6
	subbyte s7
	subbyte s8
	subbyte s9
	subbyte s10
	subbyte s11
	subbyte s12
	subbyte s13
	subbyte s14
	subbyte s15
	return
;============================================================================
; Function to perform Shift Row and Substitution
;============================================================================
fShiftRow  movff tmp, s1
	movff s1,s5
	movff s5,s9
	movff s9,s13
	movff s13,tmp

	movff tmp,s2
	movff s2,s10
	movff s10,tmp
	movff tmp,s6
	movff s6,s14
	movff s14,tmp

	movff tmp,s15
	movff s15,s11
	movff s11,s7
	movff s7,s3
	movff s3,tmp
	return
;==============================================================================
; Function to calculate GF(256) of 2W
;==============================================================================
get2	movwf	tmp
	bcf	status,c
	rlf	tmp,w
	btfss	tmp,7
	return
	xorlw	0x1b
	return
;==============================================================================
; Function to calculate GF(256) of 3W
;==============================================================================
get3	movwf	tmp
	call	get2
	xorwf	tmp,w
	return
;==============================================================================
; Function to calculate GF(256) of 9W
;==============================================================================
;get9	movwf	a1	; 1
;	call	get2
;	call	get2
;	call	get2
;	xorwf	a1,w
;	return
;==============================================================================
; Function to calculate GF(256) of BW
;==============================================================================
;getb	movwf	a1	; 1
;	call	get2	;
;	movwf	a2	; 2
;	call	get2
;	call	get2
;	xorwf	a2,w
;	xorwf	a1,w
;	return
;==============================================================================
; Function to calculate GF(256) of DW
;==============================================================================
;getd	movwf	a1	; 1
;	call	get2	;
;	call	get2
;	movwf	a2	; 4
;	call	get2
;	xorwf	a2,w
;	xorwf	a1,w
;	return
;==============================================================================
; Function to calculate GF(256) of EW
;==============================================================================
;gete	call	get2
;	movwf	a1
;	call	get2
;	movwf	a2
;	call	get2
;	xorwf	a2,w
;	xorwf	a1,w
;	return
;==============================================================================
; Subroutine to perform mix column function
;==============================================================================
;
; |                 |    |         |
; | s0  s1  s2  s3  |    | 2 1 1 3 |
; | s4  s5  s6  s7  |    | 3 2 1 1 |
; | s8  s9  s10 s11 |    | 1 3 2 1 |
; | s12 s13 s14 s15 |    | 1 1 3 2 |
; |                 |    |         |
mixcol	movf	s0,w	; row 1
	movwf	tmp1
	xorwf	s1,w
	xorwf	s2,w
	xorwf	s3,w
	movwf	tmpc

	movf	s0,w
	xorwf	s1,w
	call	get2
	xorwf	tmpc,w
	xorwf	s0,f

	movf	s1,w
	xorwf	s2,w
	call	get2
	xorwf	tmpc,w
	xorwf	s1,f

	movf	s2,w
	xorwf	s3,w
	call	get2
	xorwf	tmpc,w
	xorwf	s2,f

	movf	s3,w
	xorwf	tmp1,w
	call	get2
	xorwf	tmpc,w
	xorwf	s3,f


	movf	s4,w	; row 2
	movwf	tmp1
	xorwf	s5,w
	xorwf	s6,w
	xorwf	s7,w
	movwf	tmpc

	movf	s4,w
	xorwf	s5,w
	call	get2
	xorwf	tmpc,w
	xorwf	s4,f

	movf	s5,w
	xorwf	s6,w
	call	get2
	xorwf	tmpc,w
	xorwf	s5,f

	movf	s6,w
	xorwf	s7,w
	call	get2
	xorwf	tmpc,w
	xorwf	s6,f

	movf	s7,w
	xorwf	tmp1,w
	call	get2
	xorwf	tmpc,w
	xorwf	s7,f

	movf	s8,w	; row 3
	movwf	tmp1
	xorwf	s9,w
	xorwf	s10,w
	xorwf	s11,w
	movwf	tmpc

	movf	s8,w
	xorwf	s9,w
	call	get2
	xorwf	tmpc,w
	xorwf	s8,f

	movf	s9,w
	xorwf	s10,w
	call	get2
	xorwf	tmpc,w
	xorwf	s9,f

	movf	s10,w
	xorwf	s11,w
	call	get2
	xorwf	tmpc,w
	xorwf	s10,f

	movf	s11,w
	xorwf	tmp1,w
	call	get2
	xorwf	tmpc,w
	xorwf	s11,f

	movf	s12,w	; row 4
	movwf	tmp1
	xorwf	s13,w
	xorwf	s14,w
	xorwf	s15,w
	movwf	tmpc

	movf	s12,w
	xorwf	s13,w
	call	get2
	xorwf	tmpc,w
	xorwf	s12,f

	movf	s13,w
	xorwf	s14,w
	call	get2
	xorwf	tmpc,w
	xorwf	s13,f

	movf	s14,w
	xorwf	s15,w
	call	get2
	xorwf	tmpc,w
	xorwf	s14,f

	movf	s15,w
	xorwf	tmp1,w
	call	get2
	xorwf	tmpc,w
	xorwf	s15,f
	return
;===============================================================================
; Sbox and Inverse Sbox Lookup Table
;===============================================================================
lookup	movwf	pcl

        org     ($+0xff)&~0xff  ; force page alignment
sbox equ ($>>8)
 dt 0x63,0x7c,0x77,0x7b,0xf2,0x6b,0x6f,0xc5,0x30,0x01,0x67,0x2b,0xfe,0xd7,0xab,0x76
 dt 0xca,0x82,0xc9,0x7d,0xfa,0x59,0x47,0xf0,0xad,0xd4,0xa2,0xaf,0x9c,0xa4,0x72,0xc0
 dt 0xb7,0xfd,0x93,0x26,0x36,0x3f,0xf7,0xcc,0x34,0xa5,0xe5,0xf1,0x71,0xd8,0x31,0x15
 dt 0x04,0xc7,0x23,0xc3,0x18,0x96,0x05,0x9a,0x07,0x12,0x80,0xe2,0xeb,0x27,0xb2,0x75
 dt 0x09,0x83,0x2c,0x1a,0x1b,0x6e,0x5a,0xa0,0x52,0x3b,0xd6,0xb3,0x29,0xe3,0x2f,0x84
 dt 0x53,0xd1,0x00,0xed,0x20,0xfc,0xb1,0x5b,0x6a,0xcb,0xbe,0x39,0x4a,0x4c,0x58,0xcf
 dt 0xd0,0xef,0xaa,0xfb,0x43,0x4d,0x33,0x85,0x45,0xf9,0x02,0x7f,0x50,0x3c,0x9f,0xa8
 dt 0x51,0xa3,0x40,0x8f,0x92,0x9d,0x38,0xf5,0xbc,0xb6,0xda,0x21,0x10,0xff,0xf3,0xd2
 dt 0xcd,0x0c,0x13,0xec,0x5f,0x97,0x44,0x17,0xc4,0xa7,0x7e,0x3d,0x64,0x5d,0x19,0x73
 dt 0x60,0x81,0x4f,0xdc,0x22,0x2a,0x90,0x88,0x46,0xee,0xb8,0x14,0xde,0x5e,0x0b,0xdb
 dt 0xe0,0x32,0x3a,0x0a,0x49,0x06,0x24,0x5c,0xc2,0xd3,0xac,0x62,0x91,0x95,0xe4,0x79
 dt 0xe7,0xc8,0x37,0x6d,0x8d,0xd5,0x4e,0xa9,0x6c,0x56,0xf4,0xea,0x65,0x7a,0xae,0x08
 dt 0xba,0x78,0x25,0x2e,0x1c,0xa6,0xb4,0xc6,0xe8,0xdd,0x74,0x1f,0x4b,0xbd,0x8b,0x8a
 dt 0x70,0x3e,0xb5,0x66,0x48,0x03,0xf6,0x0e,0x61,0x35,0x57,0xb9,0x86,0xc1,0x1d,0x9e
 dt 0xe1,0xf8,0x98,0x11,0x69,0xd9,0x8e,0x94,0x9b,0x1e,0x87,0xe9,0xce,0x55,0x28,0xdf
 dt 0x8c,0xa1,0x89,0x0d,0xbf,0xe6,0x42,0x68,0x41,0x99,0x2d,0x0f,0xb0,0x54,0xbb,0x16

 END
